import 'dart:async';

import 'package:adeptarch_base/src/api/rest_api.dart' as RestAPI;
import 'package:adeptutils/adeptutils.dart';
import 'package:chopper/chopper.dart';
import 'package:http/http.dart' as http;

/// Created by krrigan on 02/27/19.

abstract class RestAPIService<I extends ChopperService,
    DataType extends Jsonable, ErrorType> implements ChopperService {
  I service;
  DataType? dataType;
  ErrorType? errorType;
  RestAPI.RestAPI restAPI;

  RestAPIService(
    this.service, {
    this.dataType,
    this.errorType,
    RestAPI.RestAPI? restAPI,
  }) : restAPI = restAPI ?? RestAPI.restAPI {
    this.restAPI.addService(service);
  }

  dynamic getAuthorization();

  void updateHttpClient(http.Client httpClient) {
    restAPI
      ..httpClient = httpClient
      ..init();
  }

  Future<Response> getSaveResponse(Future<Response> futureResponse) async {
    Response response;
    try {
      response = await futureResponse;
    } catch (e) {
      if (e is Response) {
        response = e;
      } else {
        rethrow;
      }
    }
    return response;
  }

  Future<Response<DataType>> parseResponse(
      Future<Response> futureResponse) async {
    return genericParseResponse(futureResponse, dataType);
  }

  Future<Response<List<DataType>>> parseResponseAsList(
      Future<Response> futureResponse) async {
    return genericParseResponseAsList(futureResponse, dataType);
  }

  Object? parseError(Response response) {
    Object? errorTypeResult = response.error;
    try {
      if (errorType is Jsonable) {
        errorTypeResult = response.error is String
            ? (errorType as Jsonable).fromJsonString(response.error?.toString())
            : (errorType as Jsonable)
                .fromJsonMap(response.error as Map<String, dynamic>?);
      }
    } catch (e) {
      print(e);
    }
    return errorTypeResult;
  }

  Future<Response<GenericDataType>>
      genericParseResponse<GenericDataType extends Jsonable?>(
          Future<Response> futureResponse, GenericDataType? dataType) async {
    Response response = await getSaveResponse(futureResponse);
    try {
      if (dataType != null) {
        GenericDataType? dataTypeResult = response.body is String
            ? dataType.fromJsonString(response.body) as GenericDataType?
            : dataType.fromJsonMap(response.body) as GenericDataType?;
        Object? errorTypeResult = parseError(response);
        return Response<GenericDataType>(response.base, dataTypeResult,
            error: errorTypeResult);
      }
    } catch (e) {
      String message = e.toString();
      response = Response(
          http.Response(
            response.body?.toString() ?? '',
            Jsonable.jsonParserError,
            headers: response.base.headers,
            isRedirect: response.base.isRedirect,
            persistentConnection: response.base.persistentConnection,
            reasonPhrase: response.base.reasonPhrase,
            request: response.base.request,
          ),
          response.body,
          error: message);
      print(e);
    }
    return Response<GenericDataType>(response.base, null,
        error: response.error);
  }

  Future<Response<List<GenericDataType>>>
      genericParseResponseAsList<GenericDataType extends Jsonable?>(
          Future<Response> futureResponse, GenericDataType? dataType) async {
    Response response = await getSaveResponse(futureResponse);
    try {
      if (dataType != null) {
        List<GenericDataType>? dataList = response.body is String
            ? dataType.fromJsonStringList(response.body)
                as List<GenericDataType>?
            : dataType.fromJsonList(response.body) as List<GenericDataType>?;
        Object? errorTypeResult = parseError(response);
        return Response<List<GenericDataType>>(response.base, dataList,
            error: errorTypeResult);
      }
    } catch (e) {
      String message = e.toString();
      response = Response(
          http.Response(
            response.body?.toString() ?? '',
            Jsonable.jsonParserError,
            headers: response.base.headers,
            isRedirect: response.base.isRedirect,
            persistentConnection: response.base.persistentConnection,
            reasonPhrase: response.base.reasonPhrase,
            request: response.base.request,
          ),
          response.body,
          error: message);
      print(e);
    }
    return Response<List<GenericDataType>>(response.base, null,
        error: response.error);
  }

  // @override
  // void dispose() => service.dispose(); // ChopperService no longer support .dispose()

  @override
  Type get definitionType => service.definitionType;

  @override
  ChopperClient get client => service.client;

  @override
  set client(ChopperClient client) => service.client = client;
}
