import 'package:adeptarch_base/src/resource/resource.dart';
import 'package:adeptarch_base/src/utils/adept_publish_subject.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:chopper/chopper.dart';
import 'package:adeptasync/adeptasync.dart';
import 'dart:async';

/// A generic class that can provide a resource backed by a local storage and the network.

typedef LoadFromCache<DataType> = Future<DataType> Function();
typedef OnLoadedFromCacheSuccess<DataType> = Future<Resource<DataType>>
    Function(Resource<DataType> resource);
typedef OnLoadedFromCacheFailed<DataType> = Future<Resource<DataType>> Function(
    dynamic exception);

typedef LoadFromStorage<DataType> = Future<DataType> Function();
typedef OnLoadedFromStorageSuccess<DataType> = Future<Resource<DataType>>
    Function(Resource<DataType> resource);
typedef OnLoadedFromStorageFailed<DataType> = Future<Resource<DataType>>
    Function(dynamic exception);

typedef LoadFromRESTAPI<DataType> = Future<Response<DataType>> Function();
typedef OnLoadedFromRESTAPISuccess<DataType> = Future<Resource<DataType>>
    Function({Resource<DataType>? resource, Response<DataType>? response});
typedef OnLoadedFromRESTAPIFailed<DataType> = Future<Resource<DataType>>
    Function({dynamic exception, int? code, Response<DataType>? response});

typedef OnFailed<DataType> = Future<Resource<DataType>> Function(
    dynamic exception);
typedef OnErrorManagement<DataType> = Future<Resource<DataType>>
    Function(Resource<DataType>? resource, {Response<DataType>? response});

class ResourceLoader<DataType> {
  final AsyncTask _loaderTask;

  AdeptPublishSubject<Resource<DataType?>>? publishSubject;

  LoadFromCache<DataType?>? loadFromCache;
  OnLoadedFromCacheSuccess<DataType?>? onLoadedFromCacheSuccess;
  OnLoadedFromCacheFailed<DataType?>? onLoadedFromCacheFailed;

  LoadFromStorage<DataType?>? loadFromStorage;
  OnLoadedFromStorageSuccess<DataType?>? onLoadedFromStorageSuccess;
  OnLoadedFromStorageFailed<DataType?>? onLoadedFromStorageFailed;

  LoadFromRESTAPI<DataType?>? loadFromRESTAPI;
  OnLoadedFromRESTAPISuccess<DataType?>? onLoadedFromRESTAPISuccess;
  OnLoadedFromRESTAPIFailed<DataType?>? onLoadedFromRESTAPIFailed;

  OnFailed<DataType?>? onFailed;
  OnErrorManagement<DataType?>? onErrorManagement;

  bool stopOnFirstResourceLoaded;

  ResourceLoader({
    this.publishSubject,
    this.loadFromCache,
    this.onLoadedFromCacheSuccess,
    this.onLoadedFromCacheFailed,
    this.loadFromStorage,
    this.onLoadedFromStorageSuccess,
    this.onLoadedFromStorageFailed,
    this.loadFromRESTAPI,
    this.onLoadedFromRESTAPISuccess,
    this.onLoadedFromRESTAPIFailed,
    this.onFailed,
    this.onErrorManagement,
    this.stopOnFirstResourceLoaded = false,
  }) : _loaderTask = AsyncTask();

  bool get loadMoreAfterCacheLoaded =>
      !stopOnFirstResourceLoaded &&
      (loadFromStorage != null || loadFromRESTAPI != null);

  bool get loadMoreAfterStorageLoaded =>
      !stopOnFirstResourceLoaded && loadFromRESTAPI != null;

  ///Cache handlers
  Future<Resource<DataType?>> handleLoadFromCache() async {
    try {
      if (loadFromCache == null) {
        return Resource(status: ResourceStatus.TRANSIENT);
      }
      DataType? dataLoaded = await loadFromCache?.call();
      final resource = dataLoaded == null || loadMoreAfterCacheLoaded
          ? Resource<DataType>.loading(dataLoaded)
          : Resource<DataType>.success(dataLoaded);
      return handleOnLoadedFromCacheSuccess(resource);
    } catch (e) {
      return handleOnLoadedFromCacheFailed(e);
    }
  }

  Future<Resource<DataType?>> handleOnLoadedFromCacheSuccess(
      Resource<DataType?> resource) async {
    if (onLoadedFromCacheSuccess != null) {
      return onLoadedFromCacheSuccess!(resource);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  Future<Resource<DataType?>> handleOnLoadedFromCacheFailed(
      dynamic exception) async {
    if (onLoadedFromCacheFailed != null) {
      return onLoadedFromCacheFailed!(exception);
    }
    Resource<DataType?> resource = Resource<DataType>.error(
        publishSubjectHasValue ? publishSubject?.value?.data : null,
        exception?.toString(),
        exception: exception);
    if (onErrorManagement != null) {
      resource = await onErrorManagement!(resource);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  ///Local Storage handlers
  Future<Resource<DataType?>> handleLoadFromStorage() async {
    try {
      if (loadFromStorage == null) {
        return Resource(status: ResourceStatus.TRANSIENT);
      }
      DataType? dataLoaded = await loadFromStorage?.call();
      final resource = dataLoaded == null || loadMoreAfterStorageLoaded
          ? Resource<DataType>.loading(dataLoaded)
          : Resource<DataType>.success(dataLoaded);
      return handleOnLoadedFromStorageSuccess(resource);
    } catch (e) {
      return handleOnLoadedFromStorageFailed(e);
    }
  }

  Future<Resource<DataType?>> handleOnLoadedFromStorageSuccess(
      Resource<DataType?> resource) async {
    if (onLoadedFromStorageSuccess != null) {
      return onLoadedFromStorageSuccess!(resource);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  Future<Resource<DataType?>> handleOnLoadedFromStorageFailed(
      dynamic exception) async {
    if (onLoadedFromStorageFailed != null) {
      return onLoadedFromStorageFailed!.call(exception);
    }
    Resource<DataType?> resource = Resource<DataType>.error(
        publishSubjectHasValue ? publishSubject?.value?.data : null,
        exception?.toString(),
        exception: exception);
    if (onErrorManagement != null) {
      resource = await onErrorManagement!(resource);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  ///REST API handlers
  Future<Resource<DataType?>> handleLoadFromRESTAPI() async {
    try {
      if (loadFromRESTAPI == null) {
        return Resource(status: ResourceStatus.TRANSIENT);
      }
      final Response<DataType?>? response = await loadFromRESTAPI?.call();
      if (response?.isSuccessful ?? false) {
        return handleOnLoadedFromRESTAPISuccess(response: response);
      } else {
        return handleOnLoadedFromRESTAPIFailed(
            code: response?.statusCode, response: response);
      }
    } catch (e) {
      return handleOnLoadedFromRESTAPIFailed(exception: e);
    }
  }

  Future<Resource<DataType?>> handleOnLoadedFromRESTAPISuccess(
      {Resource<DataType?>? resource, Response<DataType?>? response}) async {
    if (onLoadedFromRESTAPISuccess != null) {
      return onLoadedFromRESTAPISuccess!(
          resource: resource, response: response);
    }
    if (response?.isSuccessful ?? false) {
      return handleOnLoadedFromRESTAPISuccess(
          resource: Resource<DataType>.success(response?.body));
    }
    resource ??= Resource<DataType>.error(null, 'Empty REST API response');
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  Future<Resource<DataType?>> handleOnLoadedFromRESTAPIFailed(
      {dynamic exception, int? code, Response<DataType?>? response}) async {
    if (onLoadedFromRESTAPIFailed != null) {
      return onLoadedFromRESTAPIFailed!(
          exception: exception, code: code, response: response);
    }
    DataType? data = response?.error != null && response?.error is DataType
        ? response?.error as DataType?
        : (publishSubjectHasValue ? publishSubject?.value?.data : null);
    Resource<DataType?> resource = Resource<DataType>.error(
        data,
        exception == null && response == null
            ? 'Empty REST API response'
            : ExceptionUtils.parseException(exception).toString(),
        code: code,
        exception: exception);
    if (onErrorManagement != null) {
      resource = await onErrorManagement!(resource, response: response);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  Future<Resource<DataType?>> handleOnFailed(dynamic exception) async {
    if (onFailed != null) return onFailed!(exception);
    Resource<DataType?> resource = Resource<DataType>.error(
        publishSubjectHasValue ? publishSubject?.value?.data : null,
        exception?.toString(),
        exception: exception);
    if (onErrorManagement != null) {
      resource = await onErrorManagement!(resource);
    }
    if (publishResults) publishSubject?.post(resource);
    return resource;
  }

  bool hasResourceCompletedLoading(Resource resource) =>
      resource.status == ResourceStatus.SUCCESS ||
      resource.status == ResourceStatus.ERROR;

  ///
  /// Loads a resource from cache -> storage -> REST API.
  /// User can opt in to stopOnFirstResourceLoaded, which means when a resource
  /// is completed loaded (dataLoaded != null) then no further load will be made
  ///
  Future<Resource<DataType?>> loadResource() async {
    Resource<DataType?> resource =
        Resource<DataType?>(status: ResourceStatus.TRANSIENT);
    try {
      if (loadFromCache != null) resource = await handleLoadFromCache();
      if (hasResourceCompletedLoading(resource)) return resource;
      if (loadFromStorage != null) resource = await handleLoadFromStorage();
      if (hasResourceCompletedLoading(resource)) return resource;
      if (loadFromRESTAPI != null) resource = await handleLoadFromRESTAPI();
      if (!hasResourceCompletedLoading(resource)) {
        resource.status = ResourceStatus.SUCCESS;
        if (publishResults) publishSubject?.post(resource);
      }
    } catch (e) {
      resource = await handleOnFailed(e);
    }
    return resource;
  }

  void load() {
    _loaderTask.execute(doInBackground: ({List? values}) => loadResource());
  }

  bool get publishResults => publishSubject != null;
  bool get publishSubjectHasValue => publishSubject?.value != null;
  bool get publishSubjectValueHasData =>
      publishSubjectHasValue && publishSubject?.value?.data != null;

  AdeptPublishSubject<Resource<DataType?>> asPublishSubject() {
    publishSubject ??= AdeptPublishSubject<Resource<DataType?>>();
    if (publishSubject?.value == null) {
      publishSubject?.post(Resource<DataType>.loading(null));
    } else {
      publishSubject
          ?.post(Resource<DataType>.loading(publishSubject?.value!.data));
    }
    load();
    return publishSubject!;
  }

  Future<Resource<DataType?>> asFuture() async {
    return await loadResource();
  }

  bool isCancelled() => _loaderTask.isCanceled;
  bool isRunning() => _loaderTask.isRunning;
  void cancel() {
    if (isRunning()) _loaderTask.cancel();
  }
}
